'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

exports.default = function (babel) {
  var t = babel.types,
      transform = babel.transform;

  // Cache the bad code's loc, when exit exportDefaultDeclaration
  // inject `console.warn(...)` after `export default` statement.
  // Could not inject `console.warn(...)` just at the place where
  // the bad code located, because the bad code may would not be
  // executed after module loaded.

  var cache = null;

  function getWarnMessage(loc) {
    return 'Probably missing "yield" keyword at method calling. (' + loc.line + ':' + loc.column + ')';
  }

  function getCalleeMeta(node) {
    var callee = node.callee;

    var name = void 0;

    if (t.isIdentifier(callee)) {
      name = callee.name;
    } else if (t.isMemberExpression(callee)) {
      var obj = callee.object;
      if (t.isCallExpression(obj)) {
        name = getCalleeMeta(obj).name;
      } else {
        while (!obj.name) {
          obj = obj.object;
        }
        name = obj.name;
      }
    }

    return { name: name, loc: callee.loc.start };
  }

  function getNameFromObjectPattern(node) {
    return t.isObjectPattern(node) ? node.properties.map(function (prop) {
      if (t.isObjectPattern(prop.value)) {
        return getNameFromObjectPattern(prop.value);
      } else if (t.Identifier(prop.value)) {
        return prop.value.name;
      }
      return null;
    }) : null;
  }

  function getEffectAndActionNames(params) {
    var names = params.map(function (param) {
      if (t.isIdentifier(param)) {
        return param.name;
      }
      return getNameFromObjectPattern(param);
    });

    return (0, _lodash2.default)(names, true);
  }

  function visitCallExpression(paramNames, path) {
    if (!t.isYieldExpression(path.parent)) {
      var _getCalleeMeta = getCalleeMeta(path.node),
          name = _getCalleeMeta.name,
          loc = _getCalleeMeta.loc;
      // is effect or action call


      if (paramNames.includes(name)) {
        var filename = path.hub.file.opts.filename;

        var filepath = './' + (0, _path.relative)(process.cwd(), filename);

        console.log();
        console.log(_chalk2.default.inverse(filepath));
        console.log(getWarnMessage(loc));

        var code = (0, _babelCodeFrame2.default)(path.hub.file.code, loc.line, loc.column, { highlightCode: true });
        console.log();
        console.log(code);
        console.log();
        console.log();

        if (process.env.NODE_ENV !== 'production' && cache) {
          cache.push({
            file: filepath,
            loc: loc
          });
        }
      }
    }
  }

  function visitMethod(path) {
    var node = path.node;
    var params = void 0;
    var generator = false;

    if (t.isObjectMethod(node)) {
      generator = node.generator;
      params = node.params;
    } else if (t.isObjectProperty(node)) {
      var val = node.value;
      if (t.isFunctionExpression(val) && val.generator) {
        generator = true;
        params = val.params;
      }
    }

    if (generator) {
      var paramNames = getEffectAndActionNames(params.slice(1));
      path.get('body').traverse({
        CallExpression: visitCallExpression.bind(null, paramNames)
      });
    }
  }

  var visitor = {
    'ObjectMethod|ObjectProperty': visitMethod
  };

  if (process.env.NODE_ENV !== 'production') {
    visitor.ExportDefaultDeclaration = {
      enter: function enter() {
        cache = [];
      },
      exit: function exit(path) {
        if (cache && cache.length) {
          cache.forEach(function (_ref) {
            var file = _ref.file,
                loc = _ref.loc;

            var code = (0, _babelCodeFrame2.default)(path.hub.file.code, loc.line, loc.column);
            var raw = 'console.warn(\'' + file + ' \\n' + getWarnMessage(loc) + ' \\n\\n' + code.split('\n').join('\\n') + '\\n \')';
            var ret = transform(raw, {
              code: true,
              babelrc: false
            });
            path.insertAfter(ret.ast.program.body[0]);
          });
        }
      }
    };
  }

  return { visitor: visitor };
};

var _chalk = require('chalk');

var _chalk2 = _interopRequireDefault(_chalk);

var _path = require('path');

var _lodash = require('lodash.flatten');

var _lodash2 = _interopRequireDefault(_lodash);

var _babelCodeFrame = require('babel-code-frame');

var _babelCodeFrame2 = _interopRequireDefault(_babelCodeFrame);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

module.exports = exports['default']; /* eslint-disable no-console */